Public Class CompoundCommand
    Inherits Command
    Private m_commands As New List(Of Command)
    Private m_name As String
    Public Sub New(ByVal name As String)
        Me.m_name = name
    End Sub
    Public Overloads Overrides ReadOnly Property Name() As String
        Get
            Return m_name
        End Get
    End Property
    Public Sub Add(ByVal cmd As Command)
        m_commands.Add(cmd)
    End Sub
    Public ReadOnly Property Count() As Integer
        Get
            Return m_commands.Count
        End Get
    End Property
    Public Overrides Sub DoAction()
        For Each cmd As Command In Me.m_commands
            cmd.DoAction()
        Next
    End Sub
    Public Overrides Sub UndoAction()
        If Me.m_commands.Count = 0 Then Exit Sub
        For i As Integer = Me.m_commands.Count - 1 To 0 Step -1
            ' Must undo in reverse order!
            Dim cmd As Command = Me.m_commands(i)
            cmd.UndoAction()
        Next
    End Sub
    Public Overrides Sub RedoAction()
        For Each cmd As Command In Me.m_commands
            cmd.RedoAction()
        Next
    End Sub
    Public Overloads Overrides ReadOnly Property IsNoop() As Boolean
        Get
            For Each cmd As Command In Me.m_commands
                If Not cmd.IsNoop Then
                    Return False
                End If
            Next
            Return True
        End Get
    End Property
    Public Sub Insert(ByVal cmd As Command)
        m_commands.Insert(0, cmd)
    End Sub
End Class

