Public Class UndoManager
    Private m_stack As List(Of Command)
    Private m_pos As Integer
    Private m_max As Integer
    Private m_exec As Command = Nothing
    Private m_compound As CompoundCommand

    Public Event StateChanged()

    Public Sub New(ByVal maxHistory As Integer)
        Me.m_stack = New List(Of Command)
        Me.m_max = maxHistory
    End Sub
    Public Sub Clear()
        If m_stack.Count > 0 Then
            m_pos = 0
            m_stack.Clear()
            RaiseEvent StateChanged()
        End If
    End Sub
    Public ReadOnly Property Executing() As Command
        Get
            Return Me.m_exec
        End Get
    End Property
    Public ReadOnly Property CanUndo() As Boolean
        Get
            Return m_pos > 0
        End Get
    End Property
    Public ReadOnly Property CanRedo() As Boolean
        Get
            Return m_pos < m_stack.Count
        End Get
    End Property
    Public ReadOnly Property Current() As Command
        Get
            If m_pos >= 0 AndAlso m_pos < m_stack.Count Then
                Return m_stack(m_pos)
            End If
            Return Nothing
        End Get
    End Property
    Private Sub Add(ByVal cmd As Command)
        If m_pos < m_stack.Count Then
            m_stack.RemoveRange(m_pos, m_stack.Count - m_pos)
        End If
        m_stack.Add(cmd)
        If m_stack.Count > Me.m_max Then
            m_stack.RemoveAt(0)
        Else
            m_pos += 1
        End If
    End Sub
    Public Sub Push(ByVal cmd As Command)
        If cmd.IsNoop Then
            Return
        End If
        ' do nothing!
        If Me.m_compound IsNot Nothing Then
            Me.m_compound.Add(cmd)
        Else
            Add(cmd)
        End If
        ' Must do command after adding it to the command stack!
        Dim saved As Command = Me.m_exec
        Try
            Me.m_exec = cmd
            cmd.DoAction()
        Finally
            Me.m_exec = saved
        End Try
        RaiseEvent StateChanged()
    End Sub
    Public Sub Undo()
        If m_pos > 0 Then
            m_pos -= 1
            Dim cmd As Command = Me.Current
            If cmd IsNot Nothing Then
                Dim saved As Command = Me.m_exec
                Try
                    Me.m_exec = cmd
                    cmd.UndoAction()
                Finally
                    Me.m_exec = saved
                End Try

            End If
            RaiseEvent StateChanged()
        End If
    End Sub
    Public Sub Redo()
        If m_pos < m_stack.Count Then
            Dim cmd As Command = Me.Current
            If cmd IsNot Nothing Then
                Dim saved As Command = Me.m_exec
                Try
                    Me.m_exec = cmd
                    cmd.RedoAction()
                Finally
                    Me.m_exec = saved
                End Try
            End If
            m_pos += 1
            RaiseEvent StateChanged()
        End If
    End Sub
    Public Function Peek() As Command
        If m_pos > 0 AndAlso m_pos - 1 < m_stack.Count Then
            Return m_stack(m_pos - 1)
        End If
        Return Nothing
    End Function

    Public Sub Merge(ByVal cmd As CompoundCommand)
        ' Replace the current command without upsetting rest of the redo stack
        ' and insert the current command into this compound command.
        Dim current As Command = Peek()
        If current IsNot Nothing Then
            m_stack.Insert(m_pos - 1, cmd)
            m_stack.RemoveAt(m_pos)
            cmd.DoAction()
            cmd.Insert(current)
        Else
            Push(cmd)
        End If
    End Sub
    Public Function OpenCompoundAction(ByVal name As String) As CompoundCommand
        Me.m_compound = New CompoundCommand(name)
        Add(Me.m_compound)
        Return Me.m_compound
    End Function
    Public Sub CloseCompoundAction()
        Me.m_compound = Nothing
    End Sub
End Class

